@@ -45,8 +45,8 @@ A tiny example:
 __docformat__ = "restructuredtext en"
 
 # The format of the __version__ line is matched by a regex in setup.py
-__version__ = "4.2.1"
-__date__ = "2015/08/12"
+__version__ = "4.2.2"
+__date__ = "2017/01/29"
 
 __all__ = [
     'URIRef',
@@ -70,15 +70,29 @@ __all__ = [
 
 import sys
 assert sys.version_info >= (2, 5, 0), "rdflib requires Python 2.5 or higher"
-del sys
 
 import logging
-import __main__
-if not hasattr(__main__, '__file__'):
-    # show log messages in interactive mode
-    logging.basicConfig(level=logging.INFO)
+_interactive_mode = False
+try:
+    import __main__
+    if not hasattr(__main__, '__file__') and sys.stdout.isatty():
+        # show log messages in interactive mode
+        _interactive_mode = True
+        logging.basicConfig(level=logging.INFO)
+    del __main__
+except ImportError:
+    #Main already imported from elsewhere
+    import warnings
+    warnings.warn('__main__ already imported', ImportWarning)
+    del warnings
+
 logger = logging.getLogger(__name__)
-logger.info("RDFLib Version: %s" % __version__)
+if _interactive_mode:
+    logger.info("RDFLib Version: %s" % __version__)
+else:
+    logger.debug("RDFLib Version: %s" % __version__)
+del _interactive_mode
+del sys
 
 
 try:
