@@ -51,14 +51,15 @@ def expandTriples(terms):
         l = len(terms)
         for i, t in enumerate(terms):
             if t == ',':
-                res.append(res[i - 3])
-                res.append(res[i - 2])
+                res.extend([res[-3], res[-2]])
             elif t == ';':
-                res.append(res[i - 3])
+                if i+1 == len(terms) or terms[i+1] == ";" or terms[i+1] == ".":
+                    continue # this semicolon is spurious
+                res.append(res[-3])
             elif isinstance(t, list):
                 # BlankNodePropertyList
                 # is this bnode the object of previous triples?
-                if (i % 3) == 2:
+                if (len(res) % 3) == 2:
                     res.append(t[0])
                 # is this a single [] ?
                 if len(t) > 1:
@@ -70,6 +71,11 @@ def expandTriples(terms):
                 res += t.asList()
             elif t != '.':
                 res.append(t)
+            if DEBUG:
+                print(len(res), t)
+        if DEBUG:
+            import json
+            print(json.dumps(res, indent=2))
 
         return res
         # print res
