@@ -13,9 +13,9 @@ ORDERBY = 'ORDER BY'
 
 import re
 import collections
-import urllib.request, urllib.error, urllib.parse
+import warnings
+import contextlib
 
-# import warnings
 try:
     from SPARQLWrapper import SPARQLWrapper, XML, POST, GET, URLENCODED, POSTDIRECTLY
 except ImportError:
@@ -23,29 +23,14 @@ except ImportError:
         "SPARQLWrapper not found! SPARQL Store will not work." +
         "Install with 'easy_install SPARQLWrapper'")
 
-import sys
-if getattr(sys, 'pypy_version_info', None) is not None \
-    or sys.platform.startswith('java') \
-        or sys.version_info[:2] < (2, 6):
-    # import elementtree as etree
-    from elementtree import ElementTree
-    assert ElementTree
-else:
-    try:
-        from xml.etree import ElementTree
-        assert ElementTree
-    except ImportError:
-        from elementtree import ElementTree
-
+from rdflib.compat import etree, etree_register_namespace
 from rdflib.plugins.stores.regexmatching import NATIVE_REGEX
 
 from rdflib.store import Store
 from rdflib.query import Result
 from rdflib import Variable, Namespace, BNode, URIRef, Literal
 from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
-
-import http.client
-import urllib.parse
+from rdflib.term import Node
 
 class NSSPARQLWrapper(SPARQLWrapper):
     nsBindings = {}
@@ -75,54 +60,23 @@ class NSSPARQLWrapper(SPARQLWrapper):
         self.queryString = self.injectPrefixes(query)
 
     def injectPrefixes(self, query):
-        return '\n'.join(
-            ['\n'.join(['PREFIX %s: <%s>' % (key, val)
-                        for key, val in list(self.nsBindings.items())]),
-             query])
+        prefixes = list(self.nsBindings.items())
+        if not prefixes:
+            return query
+        return '\n'.join([
+            '\n'.join(['PREFIX %s: <%s>' % (k, v) for k, v in prefixes]),
+            '',  # separate prefixes from query with an empty line
+            query
+        ])
+
 
 BNODE_IDENT_PATTERN = re.compile('(?P<label>_\:[^\s]+)')
 SPARQL_NS = Namespace('http://www.w3.org/2005/sparql-results#')
 sparqlNsBindings = {'sparql': SPARQL_NS}
-ElementTree._namespace_map["sparql"] = SPARQL_NS
+etree_register_namespace("sparql", SPARQL_NS)
 
 
-def TraverseSPARQLResultDOM(doc, asDictionary=False):
-    """
-    Returns a generator over tuples of results
-    """
-    # namespace handling in elementtree xpath sub-set is not pretty :(
-    vars = [Variable(v.attrib["name"]) for v in doc.findall(
-            './{http://www.w3.org/2005/sparql-results#}head/' +
-            '{http://www.w3.org/2005/sparql-results#}variable')]
-    for result in doc.findall(
-            './{http://www.w3.org/2005/sparql-results#}results/' +
-            '{http://www.w3.org/2005/sparql-results#}result'):
-        currBind = {}
-        values = []
-        for binding in result.findall(
-                '{http://www.w3.org/2005/sparql-results#}binding'):
-            varVal = binding.attrib["name"]
-            var = Variable(varVal)
-            term = CastToTerm(binding.findall('*')[0])
-            values.append(term)
-            currBind[var] = term
-        if asDictionary:
-            yield currBind, vars
-        else:
-            def __locproc(values):
-                if len(values) == 1:
-                    return values[0]
-                else:
-                    return tuple(values)
-            yield __locproc(values), vars
-
-
-def localName(qname):
-    # wtf - elementtree cant do this for me
-    return qname[qname.index("}") + 1:]
-
-
-def CastToTerm(node):
+def _node_from_result(node):
     """
     Helper function that casts XML node in SPARQL results
     to appropriate rdflib term
@@ -134,8 +88,8 @@ def CastToTerm(node):
     elif node.tag == '{%s}literal' % SPARQL_NS:
         value = node.text if node.text is not None else ''
         if 'datatype' in node.attrib:
-            dT = URIRef(node.attrib['datatype'])
-            return Literal(value, datatype=dT)
+            dt = URIRef(node.attrib['datatype'])
+            return Literal(value, datatype=dt)
         elif '{http://www.w3.org/XML/1998/namespace}lang' in node.attrib:
             return Literal(value, lang=node.attrib[
                 "{http://www.w3.org/XML/1998/namespace}lang"])
@@ -145,6 +99,84 @@ def CastToTerm(node):
         raise Exception('Unknown answer type')
 
 
+def CastToTerm(node):
+    warnings.warn(
+        "Call to deprecated function CastToTerm, use _node_from_result.",
+        category=DeprecationWarning,
+    )
+    return _node_from_result(node)
+
+
+def _node_to_sparql(node):
+    if isinstance(node, BNode):
+        raise Exception(
+            "SPARQLStore does not support BNodes! "
+            "See http://www.w3.org/TR/sparql11-query/#BGPsparqlBNodes"
+        )
+    return node.n3()
+
+
+
+def _traverse_sparql_result_dom(
+        doc, as_dictionary=False, node_from_result=_node_from_result):
+    """
+    Returns a generator over tuples of results
+    """
+    # namespace handling in elementtree xpath sub-set is not pretty :(
+    vars_ = [
+        Variable(v.attrib["name"])
+        for v in doc.findall(
+            './{http://www.w3.org/2005/sparql-results#}head/'
+            '{http://www.w3.org/2005/sparql-results#}variable'
+        )
+    ]
+    for result in doc.findall(
+            './{http://www.w3.org/2005/sparql-results#}results/'
+            '{http://www.w3.org/2005/sparql-results#}result'):
+        curr_bind = {}
+        values = []
+        for binding in result.findall(
+                '{http://www.w3.org/2005/sparql-results#}binding'):
+            var_val = binding.attrib["name"]
+            var = Variable(var_val)
+            term = node_from_result(binding.findall('*')[0])
+            values.append(term)
+            curr_bind[var] = term
+        if as_dictionary:
+            yield curr_bind, vars_
+        else:
+            def __locproc(values_):
+                if len(values_) == 1:
+                    return values_[0]
+                else:
+                    return tuple(values_)
+            yield __locproc(values), vars_
+
+
+def TraverseSPARQLResultDOM(doc, asDictionary=False):
+    warnings.warn(
+        "Call to deprecated function TraverseSPARQLResultDOM, use "
+        "_traverse_sparql_result_dom instead and update asDictionary arg to "
+        "as_dictionary.",
+        category=DeprecationWarning,
+    )
+    return _traverse_sparql_result_dom(
+        doc, as_dictionary=asDictionary, node_from_result=_node_from_result)
+
+
+def _local_name(qname):
+    # wtf - elementtree cant do this for me
+    return qname[qname.index("}") + 1:]
+
+
+def localName(qname):
+    warnings.warn(
+        "Call to deprecated unused function localName, will be dropped soon.",
+        category=DeprecationWarning,
+    )
+    return _local_name(qname)
+
+
 class SPARQLStore(NSSPARQLWrapper, Store):
     """
     An RDFLib store around a SPARQL endpoint
@@ -161,13 +193,24 @@ class SPARQLStore(NSSPARQLWrapper, Store):
     Fuseki/TDB has a flag for specifying that the default graph
     is the union of all graphs (tdb:unionDefaultGraph in the Fuseki config).
 
-    .. warning:: The SPARQL Store does not support blank-nodes!
+    .. warning:: By default the SPARQL Store does not support blank-nodes!
 
-                 As blank-nodes act as variables in SPARQL queries
-                 there is no way to query for a particular blank node.
+                 As blank-nodes act as variables in SPARQL queries,
+                 there is no way to query for a particular blank node without
+                 using non-standard SPARQL extensions.
 
                  See http://www.w3.org/TR/sparql11-query/#BGPsparqlBNodes
 
+    You can make use of such extensions through the node_to_sparql and
+    node_from_result arguments. For example if you want to transform
+    BNode('0001') into "<bnode:b0001>", you can use a function like this:
+
+    >>> def my_bnode_ext(node):
+    ...    if isinstance(node, BNode):
+    ...        return '<bnode:b%s>' % node
+    ...    return _node_to_sparql(node)
+    >>> store = SPARQLStore('http://dbpedia.org/sparql',
+    ...                     node_to_sparql=my_bnode_ext)
 
     """
     formula_aware = False
@@ -178,16 +221,30 @@ class SPARQLStore(NSSPARQLWrapper, Store):
     def __init__(self,
                  endpoint=None, bNodeAsURI=False,
                  sparql11=True, context_aware=True,
+                 node_to_sparql=_node_to_sparql,
+                 node_from_result=_node_from_result,
+                 default_query_method=GET,
                  **sparqlwrapper_kwargs):
         """
         """
-        super(SPARQLStore, self).__init__(endpoint, returnFormat=XML, **sparqlwrapper_kwargs)
+        super(SPARQLStore, self).__init__(
+            endpoint, returnFormat=XML, **sparqlwrapper_kwargs)
         self.setUseKeepAlive()
         self.bNodeAsURI = bNodeAsURI
+        if bNodeAsURI:
+            warnings.warn(
+                "bNodeAsURI argument was never supported and will be dropped "
+                "in favor of node_to_sparql and node_from_result args.",
+                category=DeprecationWarning,
+            )
+        self.node_to_sparql = node_to_sparql
+        self.node_from_result = node_from_result
         self.nsBindings = {}
         self.sparql11 = sparql11
         self.context_aware = context_aware
         self.graph_aware = context_aware
+        self._timeout = None
+        self.query_method = default_query_method
 
     # Database Management Methods
     def create(self, configuration):
@@ -250,14 +307,18 @@ class SPARQLStore(NSSPARQLWrapper, Store):
             # VALUES was added to SPARQL 1.1 on 2012/07/24
             query += "\nVALUES ( %s )\n{ ( %s ) }\n"\
                 % (" ".join("?" + str(x) for x in v),
-                   " ".join(initBindings[x].n3() for x in v))
+                   " ".join(self.node_to_sparql(initBindings[x]) for x in v))
 
         self.resetQuery()
+        self.setMethod(self.query_method)
         if self._is_contextual(queryGraph):
-            self.addDefaultGraph(queryGraph)
+            self.addParameter("default-graph-uri", queryGraph)
+        self.timeout = self._timeout
         self.setQuery(query)
 
-        return Result.parse(SPARQLWrapper.query(self).response)
+        with contextlib.closing(SPARQLWrapper.query(self).response) as res:
+            return Result.parse(res)
+
 
     def triples(self, xxx_todo_changeme2, context=None):
         """
@@ -295,12 +356,6 @@ class SPARQLStore(NSSPARQLWrapper, Store):
         ``
         """
         (s, p, o) = xxx_todo_changeme2
-        if ( isinstance(s, BNode) or
-             isinstance(p, BNode) or
-             isinstance(o, BNode) ):
-            raise Exception("SPARQLStore does not support Bnodes! "
-                            "See http://www.w3.org/TR/sparql11-query/#BGPsparqlBNodes")
-
         vars = []
         if not s:
             s = Variable('s')
@@ -318,12 +373,12 @@ class SPARQLStore(NSSPARQLWrapper, Store):
         else:
             v = '*'
 
-        query = "SELECT %s WHERE { %s %s %s }" % \
-            (v, s.n3(), p.n3(), o.n3())
+        nts = self.node_to_sparql
+        query = "SELECT %s WHERE { %s %s %s }" % (v, nts(s), nts(p), nts(o))
 
         # The ORDER BY is necessary
         if hasattr(context, LIMIT) or hasattr(context, OFFSET) \
-            or hasattr(context, ORDERBY):
+                or hasattr(context, ORDERBY):
             var = None
             if isinstance(s, Variable):
                 var = s
@@ -347,12 +402,18 @@ class SPARQLStore(NSSPARQLWrapper, Store):
 
         self.resetQuery()
         if self._is_contextual(context):
-            self.addDefaultGraph(context.identifier)
+            self.addParameter("default-graph-uri", context.identifier)
+        self.timeout = self._timeout
         self.setQuery(query)
 
-        doc = ElementTree.parse(SPARQLWrapper.query(self).response)
+        with contextlib.closing(SPARQLWrapper.query(self).response) as res:
+            doc = etree.parse(res)
+
         # ElementTree.dump(doc)
-        for rt, vars in TraverseSPARQLResultDOM(doc, asDictionary=True):
+        for rt, vars in _traverse_sparql_result_dom(
+                doc,
+                as_dictionary=True,
+                node_from_result=self.node_from_result):
             yield (rt.get(s, s),
                    rt.get(p, p),
                    rt.get(o, o)), None
@@ -377,11 +438,19 @@ class SPARQLStore(NSSPARQLWrapper, Store):
             self.resetQuery()
             q = "SELECT (count(*) as ?c) WHERE {?s ?p ?o .}"
             if self._is_contextual(context):
-                self.addDefaultGraph(context.identifier)
+                self.addParameter("default-graph-uri", context.identifier)
             self.setQuery(q)
-            doc = ElementTree.parse(SPARQLWrapper.query(self).response)
+
+            with contextlib.closing(SPARQLWrapper.query(self).response) as res:
+                doc = etree.parse(res)
+
             rt, vars = next(iter(
-                TraverseSPARQLResultDOM(doc, asDictionary=True)))
+                _traverse_sparql_result_dom(
+                    doc,
+                    as_dictionary=True,
+                    node_from_result=self.node_from_result
+                )
+            ))
             return int(rt.get(Variable("c")))
 
     def contexts(self, triple=None):
@@ -401,18 +470,23 @@ class SPARQLStore(NSSPARQLWrapper, Store):
         self.resetQuery()
 
         if triple:
+            nts = self.node_to_sparql
             s, p, o = triple
-            params = ((s if s else Variable('s')).n3(),
-                      (p if p else Variable('p')).n3(),
-                      (o if o else Variable('o')).n3())
+            params = (nts(s if s else Variable('s')),
+                      nts(p if p else Variable('p')),
+                      nts(o if o else Variable('o')))
             self.setQuery('SELECT ?name WHERE { GRAPH ?name { %s %s %s }}' % params)
         else:
             self.setQuery('SELECT ?name WHERE { GRAPH ?name {} }')
 
-        doc = ElementTree.parse(SPARQLWrapper.query(self).response)
+        with contextlib.closing(SPARQLWrapper.query(self).response) as res:
+            doc = etree.parse(res)
 
-        return (rt.get(Variable("name"))
-                for rt, vars in TraverseSPARQLResultDOM(doc, asDictionary=True))
+        return (
+            rt.get(Variable("name"))
+            for rt, vars in _traverse_sparql_result_dom(
+                doc, as_dictionary=True, node_from_result=self.node_from_result)
+        )
 
     # Namespace persistence interface implementation
     def bind(self, prefix, namespace):
@@ -462,18 +536,11 @@ class SPARQLUpdateStore(SPARQLStore):
 
     For Graph objects, everything works as expected.
 
-    .. warning:: The SPARQL Update Store does not support blank-nodes!
-
-                 As blank-nodes acts as variables in SPARQL queries
-                 there is no way to query for a particular blank node.
-
-                 See http://www.w3.org/TR/sparql11-query/#BGPsparqlBNodes
-
-
+    See the :class:`SPARQLStore` base class for more information.
 
     """
 
-    where_pattern = re.compile(r"""(?P<where>WHERE\s*{)""", re.IGNORECASE)
+    where_pattern = re.compile(r"""(?P<where>WHERE\s*\{)""", re.IGNORECASE)
 
     ##################################################################
     ### Regex for injecting GRAPH blocks into updates on a context ###
@@ -525,34 +592,43 @@ class SPARQLUpdateStore(SPARQLStore):
                  queryEndpoint=None, update_endpoint=None,
                  bNodeAsURI=False, sparql11=True,
                  context_aware=True,
-                 postAsEncoded=True, autocommit=True):
+                 postAsEncoded=True, autocommit=True,
+                 **kwds
+                 ):
 
-        SPARQLStore.__init__(self,
-                             queryEndpoint, bNodeAsURI, sparql11, context_aware, updateEndpoint=update_endpoint)
+        SPARQLStore.__init__(
+            self,
+            queryEndpoint,
+            bNodeAsURI,
+            sparql11,
+            context_aware,
+            updateEndpoint=update_endpoint,
+            **kwds
+        )
 
         self.postAsEncoded = postAsEncoded
         self.autocommit = autocommit
         self._edits = None
 
-    def query(self,*args, **kwargs):
+    def query(self, *args, **kwargs):
         if not self.autocommit:
             self.commit()
-        return SPARQLStore.query(self,*args, **kwargs)
-	
-    def triples(self,*args, **kwargs):
+        return SPARQLStore.query(self, *args, **kwargs)
+
+    def triples(self, *args, **kwargs):
         if not self.autocommit:
             self.commit()
-        return SPARQLStore.triples(self,*args, **kwargs)
-	
-    def contexts(self,*args, **kwargs):
+        return SPARQLStore.triples(self, *args, **kwargs)
+
+    def contexts(self, *args, **kwargs):
         if not self.autocommit:
             self.commit()
-        return SPARQLStore.contexts(self,*args, **kwargs)
-	
-    def __len__(self,*args, **kwargs):
+        return SPARQLStore.contexts(self, *args, **kwargs)
+
+    def __len__(self, *args, **kwargs):
         if not self.autocommit:
             self.commit()
-        return SPARQLStore.__len__(self,*args, **kwargs)
+        return SPARQLStore.__len__(self, *args, **kwargs)
 
     def open(self, configuration, create=False):
         """
@@ -576,7 +652,7 @@ class SPARQLUpdateStore(SPARQLStore):
             self.updateEndpoint = self.endpoint
 
     def _transaction(self):
-        if self._edits == None:
+        if self._edits is None:
             self._edits = []
         return self._edits
 
@@ -589,20 +665,19 @@ class SPARQLUpdateStore(SPARQLStore):
     update_endpoint = property(
         __get_update_endpoint,
         __set_update_endpoint,
-        doc='the HTTP URL for the Update endpoint, typically' +
+        doc='the HTTP URL for the Update endpoint, typically '
             'something like http://server/dataset/update')
 
     # Transactional interfaces
     def commit(self):
-        """ add(), addN(), and remove() are transactional to reduce overhead of many small edits. 
-            Read and update() calls will automatically commit any outstanding edits. 
-            This should behave as expected most of the time, except that alternating writes 
+        """ add(), addN(), and remove() are transactional to reduce overhead of many small edits.
+            Read and update() calls will automatically commit any outstanding edits.
+            This should behave as expected most of the time, except that alternating writes
             and reads can degenerate to the original call-per-triple situation that originally existed.
         """
         if self._edits and len(self._edits) > 0:
-            r = self._do_update('\n;\n'.join(self._edits))
+            self._do_update('\n;\n'.join(self._edits))
             self._edits = None
-            return r
 
     def rollback(self):
         self._edits = None
@@ -616,17 +691,11 @@ class SPARQLUpdateStore(SPARQLStore):
         assert not quoted
         (subject, predicate, obj) = spo
 
-        if ( isinstance(subject, BNode) or
-             isinstance(predicate, BNode) or
-             isinstance(obj, BNode) ):
-            raise Exception("SPARQLStore does not support Bnodes! "
-                            "See http://www.w3.org/TR/sparql11-query/#BGPsparqlBNodes")
-
-
-        triple = "%s %s %s ." % (subject.n3(), predicate.n3(), obj.n3())
+        nts = self.node_to_sparql
+        triple = "%s %s %s ." % (nts(subject), nts(predicate), nts(obj))
         if self._is_contextual(context):
             q = "INSERT DATA { GRAPH %s { %s } }" % (
-                context.identifier.n3(), triple)
+                nts(context.identifier), triple)
         else:
             q = "INSERT DATA { %s }" % triple
         self._transaction().append(q)
@@ -642,9 +711,15 @@ class SPARQLUpdateStore(SPARQLStore):
         for subject, predicate, obj, context in quads:
             contexts[context].append((subject,predicate,obj))
         data = []
+        nts = self.node_to_sparql
         for context in contexts:
-            triples = ["%s %s %s ." % (x[0].n3(), x[1].n3(), x[2].n3()) for x in contexts[context]]
-            data.append("INSERT DATA { GRAPH <%s> { %s } }\n" % (context.identifier, '\n'.join(triples)))
+            triples = [
+                "%s %s %s ." % (
+                    nts(subject), nts(predicate), nts(obj)
+                ) for subject, predicate, obj in contexts[context]
+            ]
+            data.append("INSERT DATA { GRAPH %s { %s } }\n" % (
+                nts(context.identifier), '\n'.join(triples)))
         self._transaction().extend(data)
         if self.autocommit:
             self.commit()
@@ -662,25 +737,36 @@ class SPARQLUpdateStore(SPARQLStore):
         if not obj:
             obj = Variable("O")
 
-        triple = "%s %s %s ." % (subject.n3(), predicate.n3(), obj.n3())
+        nts = self.node_to_sparql
+        triple = "%s %s %s ." % (nts(subject), nts(predicate), nts(obj))
         if self._is_contextual(context):
+            cid = nts(context.identifier)
             q = "DELETE { GRAPH %s { %s } } WHERE { GRAPH %s { %s } }" % (
-                context.identifier.n3(), triple,
-                context.identifier.n3(), triple)
+                cid, triple,
+                cid, triple)
         else:
             q = "DELETE { %s } WHERE { %s } " % (triple, triple)
         self._transaction().append(q)
         if self.autocommit:
             self.commit()
 
+    def setTimeout(self, timeout):
+        self._timeout = int(timeout)
+
     def _do_update(self, update):
         self.resetQuery()
         self.setQuery(update)
         self.setMethod(POST)
+        self.timeout = self._timeout
         self.setRequestMethod(URLENCODED if self.postAsEncoded else POSTDIRECTLY)
 
         result = SPARQLWrapper.query(self)
-        return result
+
+        # we must read (and discard) the whole response
+        # otherwise the network socket buffer will at some point be "full"
+        # and we will block
+        with contextlib.closing(result.response) as res:
+            res.read()
 
     def update(self, query,
                initNs={},
@@ -739,7 +825,7 @@ class SPARQLUpdateStore(SPARQLStore):
             v = list(initBindings)
             values = "\nVALUES ( %s )\n{ ( %s ) }\n"\
                 % (" ".join("?" + str(x) for x in v),
-                   " ".join(initBindings[x].n3() for x in v))
+                   " ".join(self.node_to_sparql(initBindings[x]) for x in v))
 
             query = self.where_pattern.sub("WHERE { " + values, query)
 
@@ -755,7 +841,11 @@ class SPARQLUpdateStore(SPARQLStore):
             is converted into
             "INSERT DATA { GRAPH <urn:graph> { <urn:michel> <urn:likes> <urn:pizza> } }"
         """
-        graph_block_open = " GRAPH <%s> {" % query_graph
+        if isinstance(query_graph, Node):
+            query_graph = self.node_to_sparql(query_graph)
+        else:
+            query_graph = '<%s>' % query_graph
+        graph_block_open = " GRAPH %s {" % query_graph
         graph_block_close = "} "
 
         # SPARQL Update supports the following operations:
@@ -802,7 +892,8 @@ class SPARQLUpdateStore(SPARQLStore):
         if not self.graph_aware:
             Store.add_graph(self, graph)
         elif graph.identifier != DATASET_DEFAULT_GRAPH_ID:
-            self.update("CREATE GRAPH <%s>" % graph.identifier)
+            self.update(
+                "CREATE GRAPH %s" % self.node_to_sparql(graph.identifier))
 
     def remove_graph(self, graph):
         if not self.graph_aware:
@@ -810,4 +901,5 @@ class SPARQLUpdateStore(SPARQLStore):
         elif graph.identifier == DATASET_DEFAULT_GRAPH_ID:
             self.update("DROP DEFAULT")
         else:
-            self.update("DROP GRAPH <%s>" % graph.identifier)
+            self.update(
+                "DROP GRAPH %s" % self.node_to_sparql(graph.identifier))
